from .rotor import Rotor
from .reflector import Reflector
from .plugboard import Plugboard

UPPERCASE_LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

class Enigma:
    """
    Represents an Enigma machine to be used for encrypting messages.
    """

    def __init__(self, reflector:Reflector, left_rotor:Rotor, middle_rotor:Rotor, right_rotor:Rotor, plugboard:Plugboard, rotor_positions:str="1 1 1", ring_positions:str="1 1 1"):

        self.reflector = reflector
        self.left_rotor = left_rotor
        self.middle_rotor = middle_rotor
        self.right_rotor = right_rotor
        self.plugboard = plugboard

        self.rotor_positions = rotor_positions.split()
        assert len(self.rotor_positions)==3
        self.ring_positions = ring_positions.split()
        assert len(self.ring_positions)==3

        self.left_rotor.rotor_position = int(self.rotor_positions[0])
        self.middle_rotor.rotor_position = int(self.rotor_positions[1])
        self.right_rotor.rotor_position = int(self.rotor_positions[2])

        self.left_rotor.set_ring_position(int(self.ring_positions[0]))
        self.middle_rotor.set_ring_position(int(self.ring_positions[1]))
        self.right_rotor.set_ring_position(int(self.ring_positions[2]))

    @classmethod
    def from_dict(cls, configs: dict):
        """
        Alternative constructor for configs which have been 
        exported by export_settings method
        """
        return cls(configs['reflector'],configs['left_rotor'],configs['middle_rotor'],configs['right_rotor'],
                Plugboard(configs['plug_settings']),configs['rotor_positions'], configs['ring_positions'],)
        

    def __setattr__(self, name: str, value) -> None:
        self.__dict__[name] = value

    def rotate(self) -> None:
        """
        Rotates the three rotors according if a rotor is in turnover position.
        """
        if self.middle_rotor.is_in_turnover_pos():
            self.middle_rotor.notch()
            self.left_rotor.notch()
        elif self.right_rotor.is_in_turnover_pos():
            self.middle_rotor.notch()
        self.right_rotor.notch()

    def encipher(self, plain_text:str)->str:
        """
        Encrypts a message by passing each character given the current enigma 
        setup through the plugboard, rotors, reflector and back.
        """
        plain_text = plain_text.upper()
        encoded_text = ""
        for char in plain_text:
            if char in [" ", ",",".",":",";","-","_","=","?","%","&","(",")","[","]"]:
                encoded_text += char
                continue
            assert char in UPPERCASE_LETTERS

            self.rotate()

            temp = self.plugboard.map_plug(char)
            temp = self.right_rotor.encipher_forward(temp)
            temp = self.middle_rotor.encipher_forward(temp)
            temp = self.left_rotor.encipher_forward(temp)
            temp = self.reflector.encipher(temp)
            temp = self.left_rotor.encipher_backwards(temp)
            temp = self.middle_rotor.encipher_backwards(temp)
            temp = self.right_rotor.encipher_backwards(temp)
            temp = self.plugboard.map_plug(temp)
            
            encoded_text += temp
        return encoded_text
    
    def export_settings(self) -> dict:
        """
        Export the instance of the enigma machine as dictionary.
        Rotor and ring positions, as well as the plugboard settings are persisted.
        To be used with from_dict method.
        """
        return {
        'reflector': self.reflector,
        'left_rotor': self.left_rotor,
        'middle_rotor': self.middle_rotor,
        'right_rotor': self.right_rotor,
        'rotor_positions': f'{self.left_rotor.rotor_position} {self.middle_rotor.rotor_position} {self.right_rotor.rotor_position}',
        'ring_positions': f'{self.left_rotor.ring_position} {self.middle_rotor.ring_position} {self.right_rotor.ring_position}',
        'plug_settings': f'{self.plugboard}'
        }

    def __str__(self):
        """
        Pretty display.
        """
        return f"""reflector: '{self.reflector}',
        left_rotor: '{self.left_rotor}',
        middle_rotor: '{self.middle_rotor}',
        right_rotor: '{self.right_rotor}',
        rotor_positions: '{self.left_rotor.rotor_position} {self.middle_rotor.rotor_position} {self.right_rotor.rotor_position}',
        ring_positions: '{self.left_rotor.ring_position} {self.middle_rotor.ring_position} {self.right_rotor.ring_position}',
        plug_settings: '{self.plugboard}'"""


UPPERCASE_LETTERS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

class Reflector:
    """
    Represents a reflector which directs current back through thr rotors
    """

    def __init__(self, wiring:str=None, name:str=None):
        if wiring != None:
            self.wiring = wiring
        else:
            self.wiring =  UPPERCASE_LETTERS
        self.name = name

    
    def __setattr__(self, name: str, value) -> None:
        self.__dict__[name] = value

    def __eq__(self, reflector) -> bool:
        return self.name == reflector.name
    
    def __str__(self) -> str:
        return self.name    
    
    def __len__(self):
        return len(self.name)

    def encipher(self, key: str)->str:
        """
        Simulates the passing of a character through the reflector.
        """
        assert type(key) == str
        assert len(key) == 1
        key = key.upper()
        assert key in UPPERCASE_LETTERS
        index = (ord(key) - ord("A"))
        letter = self.wiring[index]
        return letter



REFLECTOR_A = Reflector(wiring="EJMZALYXVBWFCRQUONTSPIKHGD", name="A")
REFLECTOR_B = Reflector(wiring="YRUHQSLDPXNGOKMIEBFZCWVJAT", name="B")
REFLECTOR_C = Reflector(wiring="FVPJIAOYEDRZXWGCTKUQSBNMHL", name="C")



from ..src.reflector import *
import pytest

def test_reflectorA_encipherZ():
    x = REFLECTOR_A.encipher(key="z")
    assert x == 'D'

def test_reflectorB_encipherZ():
    x = REFLECTOR_B.encipher(key="z")
    assert x == 'T'

def test_reflectorC_encipherZ():
    x = REFLECTOR_C.encipher(key="z")
    assert x == 'L'

def test_reflector_encipherInt():
    with pytest.raises(AssertionError):
        REFLECTOR_A.encipher(9)

def test_reflector_encipherTwoStr():
    with pytest.raises(AssertionError):
        REFLECTOR_A.encipher('AA')
    



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="enigmator",
    version="1.0.1",
    description="Package for decrypting messages with an enigma machine",
    package_dir={"": "enigma"},
    packages=find_packages(where="enigma"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/gitstetter/enigma',
    author="Andreas Fussstetter",
    author_email="a.fussstetter@hotmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
    ],
    extras_require={
        "dev": ["pytest>=5.4.1", "twine>=4.0.2"],
    },
    python_requires=">=3.7.7",
)
